package script

import (
	"github.com/d5/tengo/compiler/token"
	"github.com/d5/tengo/objects"
)

// Tengo object type that wraps a read-only map[string]int64, without
// requiring us to go through a map[string]interface{}.
type intMap map[string]int64

func (m intMap) String() string {
	return "<intMap>"
}

func (m intMap) TypeName() string {
	return "int-map"
}

func (m intMap) Copy() objects.Object {
	return m
}

func (m intMap) IsFalsy() bool {
	return len(m) == 0
}

func (m intMap) Equals(o objects.Object) bool {
	return false
}

func (m intMap) BinaryOp(op token.Token, rhs objects.Object) (objects.Object, error) {
	return nil, objects.ErrInvalidOperator
}

func (m intMap) IndexGet(index objects.Object) (objects.Object, error) {
	indexStr, ok := index.(*objects.String)
	if !ok {
		return nil, objects.ErrInvalidIndexType
	}
	value, ok := m[indexStr.Value]
	if !ok {
		return objects.UndefinedValue, nil
	}
	return &objects.Int{Value: value}, nil
}

func (m intMap) Iterate() objects.Iterator {
	return newIntMapIterator(m)
}

// We need to make a copy of the map to iterate on it with this API.
// Note that in the iterator, 'idx' maps to the *next* item.
type intMapIterator struct {
	arr []keyIntPair
	idx int
}

type keyIntPair struct {
	key   string
	value int64
}

func newIntMapIterator(m intMap) *intMapIterator {
	arr := make([]keyIntPair, 0, len(m))
	for k, v := range m {
		arr = append(arr, keyIntPair{key: k, value: v})
	}
	return &intMapIterator{arr: arr}
}

func (i *intMapIterator) String() string {
	return "<intMapIterator>"
}

func (i *intMapIterator) TypeName() string {
	return "int-map-iterator"
}

func (i *intMapIterator) Copy() objects.Object {
	return i
}

func (i *intMapIterator) IsFalsy() bool {
	return len(i.arr) == 0
}

func (i *intMapIterator) Equals(o objects.Object) bool {
	return false
}

func (i *intMapIterator) BinaryOp(op token.Token, rhs objects.Object) (objects.Object, error) {
	return nil, objects.ErrInvalidOperator
}

func (i *intMapIterator) Next() bool {
	i.idx++
	return i.idx <= len(i.arr)
}

func (i *intMapIterator) Key() objects.Object {
	return &objects.String{Value: i.arr[i.idx-1].key}
}

func (i *intMapIterator) Value() objects.Object {
	return &objects.Int{Value: i.arr[i.idx-1].value}
}
