package server

import (
	"context"
	"log"
	"time"

	"github.com/golang/protobuf/ptypes/empty"
	"google.golang.org/grpc/codes"
	"google.golang.org/grpc/status"

	"git.autistici.org/ai3/tools/iprep/db"
	ippb "git.autistici.org/ai3/tools/iprep/proto"
	"git.autistici.org/ai3/tools/iprep/script"
)

type Server struct {
	*script.Manager
	horizon time.Duration
	db      db.DB
	stop    chan struct{}
}

var (
	emptyResponse = new(empty.Empty)

	defaultHorizon = -6 * time.Hour
	dataMaxAge     = 168 * time.Hour // one week

	defaultScript = `score = 0`
)

func New(dbPath, scriptPath string) (*Server, error) {
	scriptMgr, err := script.NewManager(scriptPath, defaultScript)
	if err != nil {
		return nil, err
	}

	database, err := db.Open(dbPath)
	if err != nil {
		return nil, err
	}

	s := &Server{
		Manager: scriptMgr,
		horizon: defaultHorizon,
		db:      database,
		stop:    make(chan struct{}),
	}
	go s.cleaner()
	return s, nil
}

func (s *Server) Close() {
	close(s.stop)
	s.db.Close()
}

func (s *Server) Submit(ctx context.Context, req *ippb.SubmitRequest) (*empty.Empty, error) {
	// Reduce the request to a single Aggregate.
	var aggr *ippb.Aggregate
	for _, a := range req.Aggregates {
		if aggr == nil {
			aggr = a
		} else {
			aggr = aggr.Merge(a)
		}
	}
	for _, e := range req.Events {
		if aggr == nil {
			aggr = new(ippb.Aggregate)
		}
		aggr.AddEvent(e)
	}

	// Load it into the database if it's not empty.
	if len(aggr.ByType) == 0 {
		return nil, status.Errorf(codes.InvalidArgument, "empty input")
	}
	err := s.db.AddAggregate(aggr)

	return emptyResponse, err
}

func (s *Server) GetScore(ctx context.Context, req *ippb.GetScoreRequest) (*ippb.GetScoreResponse, error) {
	counts, err := s.db.ScanIP(time.Now().Add(s.horizon), req.Ip)
	if err != nil {
		return nil, status.Errorf(codes.Unavailable, "%v", err)
	}

	score, err := s.Script().RunIP(ctx, req.Ip, counts, s.horizon.Seconds(), nil)
	if err != nil {
		return nil, status.Errorf(codes.Internal, "%v", err)
	}

	return &ippb.GetScoreResponse{
		Ip:    req.Ip,
		Score: float32(score),
	}, nil
}

func (s *Server) cleaner() {
	t := time.NewTicker(6 * time.Hour)
	for {
		select {
		case <-s.stop:
			return
		case <-t.C:
			if err := s.db.WipeOldData(dataMaxAge); err != nil {
				log.Printf("error cleaning up database: %v", err)
			}
		}
	}
}
